@@ -7,8 +7,13 @@
 {-# LANGUAGE FlexibleContexts #-}
 {-# LANGUAGE ScopedTypeVariables #-}
 {-# LANGUAGE UndecidableInstances #-}
-#if defined(__GLASGOW_HASKELL__) && __GLASGOW_HASKELL__ >= 704
+#if defined(__GLASGOW_HASKELL__)
+# if __GLASGOW_HASKELL__ >= 704
 {-# LANGUAGE Trustworthy #-}
+# endif
+# if __GLASGOW_HASKELL__ >= 708
+{-# LANGUAGE EmptyCase #-}
+# endif
 #endif
 #ifndef MIN_VERSION_base
 #define MIN_VERSION_base(x,y,z) 1
@@ -50,6 +55,7 @@ module Data.Bytes.Serial
 
 import Control.Applicative
 import Control.Monad
+import qualified Control.Monad.Fail as MonadFail
 import qualified Data.Foldable as F
 import Data.Bytes.Get
 import Data.Bytes.Put
@@ -274,7 +280,7 @@ restore :: forall m a. (MonadGet m, Storable a) => m a
 restore = do
   let required = sizeOf (undefined :: a)
   PS fp o n <- getByteString required
-  unless (n >= required) $ fail "restore: Required more bytes"
+  unless (n >= required) $ MonadFail.fail "restore: Required more bytes"
   return $ unsafePerformIO $ withForeignPtr fp $ \p -> peekByteOff p o
 
 instance Serial Double where
@@ -337,7 +343,7 @@ instance Serial Sci.Scientific where
 
 instance Serial Void where
   serialize = absurd
-  deserialize = fail "I looked into the void."
+  deserialize = MonadFail.fail "I looked into the void."
 
 instance Serial ISet.IntSet where
   serialize = serialize . ISet.toAscList
@@ -594,8 +600,13 @@ instance GSerial U1 where
   gdeserialize = return U1
 
 instance GSerial V1 where
-  gserialize _ = fail "I looked into the void."
-  gdeserialize = fail "I looked into the void."
+  gserialize x =
+#if __GLASGOW_HASKELL__ >= 708
+    case x of {}
+#else
+    x `seq` error "I looked into the void."
+#endif
+  gdeserialize = MonadFail.fail "I looked into the void."
 
 instance (GSerial f, GSerial g) => GSerial (f :*: g) where
   gserialize (f :*: g) = do
@@ -609,7 +620,7 @@ instance (GSerial f, GSerial g) => GSerial (f :+: g) where
   gdeserialize = getWord8 >>= \a -> case a of
     0 -> liftM L1 gdeserialize
     1 -> liftM R1 gdeserialize
-    _ -> fail "Missing case"
+    _ -> MonadFail.fail "Missing case"
 
 instance GSerial f => GSerial (M1 i c f) where
   gserialize (M1 x) = gserialize x
@@ -772,8 +783,13 @@ instance GSerial1 U1 where
   gdeserializeWith _  = return U1
 
 instance GSerial1 V1 where
-  gserializeWith _   = fail "I looked into the void."
-  gdeserializeWith _ = fail "I looked into the void."
+  gserializeWith _ x =
+#if __GLASGOW_HASKELL__ >= 708
+    case x of {}
+#else
+    x `seq` error "I looked into the void."
+#endif
+  gdeserializeWith _ = MonadFail.fail "I looked into the void."
 
 instance (GSerial1 f, GSerial1 g) => GSerial1 (f :*: g) where
   gserializeWith f (a :*: b) = gserializeWith f a >> gserializeWith f b
@@ -785,7 +801,7 @@ instance (GSerial1 f, GSerial1 g) => GSerial1 (f :+: g) where
   gdeserializeWith m = getWord8 >>= \a -> case a of
     0 -> liftM L1 (gdeserializeWith m)
     1 -> liftM R1 (gdeserializeWith m)
-    _ -> fail "Missing case"
+    _ -> MonadFail.fail "Missing case"
 
 instance (Serial1 f, GSerial1 g) => GSerial1 (f :.: g) where
   gserializeWith f (Comp1 m) = serializeWith (gserializeWith f) m
@@ -821,7 +837,7 @@ instance Serial2 Either where
   deserializeWith2 m n = getWord8 >>= \a -> case a of
     0 -> liftM Left m
     1 -> liftM Right n
-    _ -> fail "Missing case"
+    _ -> MonadFail.fail "Missing case"
 
 instance Serial2 (,) where
   serializeWith2 f g (a, b) = f a >> g b
